
#include <WebServer.h>
#include <HTTPUpdateServer.h>

// send
// 98F7FE6400EA 38 7DFF 38 = read snapshot
// 98F7FEXX00E6 37 0283LOHIYYFF 37 = set power
// receive
// 83 83 FD 02 55 51 18 02 97 FF sync
// 83 83 FE 3C 00 E6 87 5A FF ack powerset

//Wifi password
const char* SSID = "FSM";
const char* PASSWORD = "0101010101";

String Shelly_IP = "192.168.178.47";

//Other
const int VEBUS_RXD1=16, VEBUS_TXD1=17, VEBUS_DE=4;  //Victron Multiplus VE.bus RS485 gpio pins

//other variables:
char rxbuf[256];                    //used to store last serial bytes received from Multiplus, only a part of a frame
char frbuf[256];                    //assembles one complete frame received by Multiplus
char txbuf1[64];                    //buffer for assembling bare command towards Multiplus (without replacements or checksum)
char txbuf2[64];                    //Multiplus output buffer containing the final command towards Multiplus, including replacements and checksum
byte rxnum;                    
byte frp = 0;                       //Pointer into Multiplus framebuffer frbuf[] to store received frames.
byte frameNr = 0;                   //Last frame number received from Multiplus. Own command has be be sent with frameNr+1, otherwise it will be ignored by Multiplus.

int meterPower = 0;                 //result of newest power measurement
short essPower = 0;
int reqPower = 0;

WebServer server(80);
HTTPUpdateServer httpUpdater;


int prepareESScommand(char *outbuf, short power, byte desiredFrameNr)
{
  byte j=0;
  outbuf[j++] = 0x98;           //MK3 interface to Multiplus
  outbuf[j++] = 0xf7;           //MK3 interface to Multiplus
  outbuf[j++] = 0xfe;           //data frame
  outbuf[j++] = desiredFrameNr;
  outbuf[j++] = 0x00;           //our own ID
  outbuf[j++] = 0xe6;           //our own ID
  outbuf[j++] = 0x37;           //CommandWriteViaID
  outbuf[j++] = 0x02;           //Flags, 0x02=RAMvar and no EEPROM
  outbuf[j++] = 0x83;           //ID = address of ESS power in assistand memory
  outbuf[j++] = (power & 0xFF); //Lo value of power (positive = into grid, negative = from grid)
  outbuf[j++] = (power >> 8);   //Hi value of power (positive = into grid, negative = from grid)
  return j;
}

int commandReplaceFAtoFF(char *outbuf, char *inbuf, int inlength)
{
  int j=0;
  //copy over the first 4 bytes of command, as there is no replacement
  for (int i = 0; i < 4; i++)
  {
    outbuf[j++] = inbuf[i];
  } 
  //starting from 5th byte, replace 0xFA..FF with double-byte character
  for (int i = 4; i < inlength; i++)
  {
    byte c = inbuf[i];
    if (c >= 0xFA)
    {
      outbuf[j++] = 0xFA;
      outbuf[j++] = 0x70 | (c & 0x0F);
    }
    else
    {
      outbuf[j++] = c;    //no replacement
    }
  }
  return j;   //new length of output frame
}

int appendChecksum(char *buf, int inlength)
{
  int j=0;
  //calculate checksum starting from 3rd byte
  byte cs=1;
  for (int i = 2; i < inlength; i++)
  {
    cs -= buf[i];
  }
  j = inlength;
  if (cs >= 0xFA)
  {
    buf[j++] = 0xFA;
    buf[j++] = (cs-0xFA);
  }
  else
  {
    buf[j++] = cs;
  }
  buf[j++] = 0xFF;  //append End Of Frame symbol
  return j;   //new length of output frame
}

bool autozero;
bool syncrxed;
bool gotmsg;
unsigned long synctime;
unsigned long getpwrtime;

void sendmsg()
{
  //send ESS command
  int len = prepareESScommand(txbuf1, essPower, (frameNr+1) & 0x7F);
  len = commandReplaceFAtoFF(txbuf2, txbuf1, len);
  len = appendChecksum(txbuf2, len);
  //write command into Multiplus :-)
  digitalWrite(VEBUS_DE,HIGH);      //set RS485 direction to write
  Serial1.write(txbuf2, len); //write command bytes to UART
  Serial1.flush(); //
  digitalWrite(VEBUS_DE,LOW);   //set RS485 direction to read
}

void multiplusCommandHandling()
{
  //Check for new bytes on UART
  int nr = Serial1.available();
  if (nr > 0)
  {
    // Serial.readStringUntil(0xFF)
    Serial1.read(rxbuf, nr);    //read all available bytes
    for (int n=0; n<nr; n++)          //loop over all new bytes
    {
      char c = rxbuf[n];       //read one byte
      frbuf[frp++] = c;   //store into framebuffer
      if (c==0x55) synctime = millis(); 
      if (c==0xFF)        //in case current byte was EndOfFrame, interprete frame
      {
        //Serial.println(frp);
        if ((frbuf[2] == 0xFD) && (frbuf[4] == 0x55))  //if it was a sync frame:
        {
          frameNr = frbuf[3];
          if (n == nr-1)       //if new ESSpower should be written and 0xFF (end of frame) was last received character:
          {
            Serial.println("Send OK");
            syncrxed = true;
          }
        }
        if ((frbuf[0] == 0x83) && (frbuf[1] == 0x83) && (frbuf[2] == 0xFE) && (frbuf[4] == 0x00) && (frbuf[5] == 0xE6) && (frbuf[6] == 0x87))  //if it is a (positive) response to an ESS write
        {
          Serial.println("Ack");
        }
        rxnum = frp;
        frp = 0;
      }
    }
  }
}

void setup() 
{
  //Setup SerialMonitor for debug information
  Serial.begin(115200);
  //Setup Serial port for VE.Bus RS485 to Multiplus
  Serial1.begin(256000, SERIAL_8N1, VEBUS_RXD1, VEBUS_TXD1);
  pinMode(VEBUS_DE, OUTPUT);    //RS485 RE/DE direction pin for UART1
  digitalWrite(VEBUS_DE,LOW);   //set RS485 direction to read
  //Init webserver
  WiFi.mode(WIFI_STA);
  WiFi.begin(SSID, PASSWORD);
  while (WiFi.status() != WL_CONNECTED) 
  {
    delay(500);
    Serial.print(".");
  }
  httpUpdater.setup(&server,"/up"); 
  server.on("/", handleRoot);
  server.begin();
  Serial.println(" connected to WiFi");
  Serial.println(WiFi.localIP());
}

void loop() 
{
  multiplusCommandHandling(); //Multiplus VEbus: Process VEbus information and execute command-sending state machine

  if (syncrxed) 
  {
    if (gotmsg)
    {
      if (millis() > synctime + 10) 
      {
        syncrxed = false;
        essPower = short(reqPower);
        sendmsg();
        gotmsg = false;
      }
    }
    else syncrxed = false;
  }
  
  if (millis() > getpwrtime)
  {
    getpwrtime = millis() + 5000;
    meterPower = http_getpwr("http://" + Shelly_IP + "/status");
    if (meterPower == 12345) meterPower = 0;
    Serial.print("Shelly Power "); Serial.println(meterPower);

    if (autozero)
    {   
      // meter 200, req 0 -> invert 200
      // meter 200 , req 100 -> invert 300
      // meter 200 , req -100 -> invert 100
      // meter -200, req 0 -> charge 200
      // meter -200, req -100 -> charge 300
      // meter -200, req 100 -> charge 100
      
      reqPower =  0.9 * (meterPower + reqPower);
      // invert, positive values
      if (reqPower > 50)
      { 
        if (reqPower > 500) reqPower = 500;
        else                reqPower -= 25;
      }
      // charge, negative values
      else if (reqPower < -50)
      { 
        if (reqPower < -500) reqPower = -500;
        else                 reqPower += 25;
      }
      // do zero
      else reqPower = 0;

      syncrxed = false;
    }
    gotmsg = true; // make sure no timeout
  }
  
  server.handleClient(); 
}
